#!/usr/bin/env python3

# Deskflow -- mouse and keyboard sharing utility
# Copyright (C) 2024 Symless Ltd.
#
# This package is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# found in the file LICENSE that should have accompanied this file.
#
# This package is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import lib.env as env

env.ensure_in_venv(__file__)

import argparse, sys
import lib.fs as fs
from clang_format import clang_format  # type: ignore

include_files = [
    "*.h",
    "*.c",
    "*.hpp",
    "*.cpp",
]

dirs = ["src"]


def main():
    """
    Cross-platform equivalent of using find and xargs with clang-format.
    Lints by performing a dry run (--dry-run) which fails when formatting is needed.
    """
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-f",
        "--format",
        action="store_true",
        help="In-place format all files",
    )
    args = parser.parse_args()

    cmd_args = ["-i"] if args.format else ["--dry-run", "--Werror"]
    files_recursive = fs.find_files(dirs, include_files)

    if args.format:
        print("Formatting files with Clang formatter:")
    else:
        print("Checking files with Clang formatter:")

    for file in files_recursive:
        print(file)

    if files_recursive:
        sys.argv = [""] + cmd_args + files_recursive
        result = clang_format()
        if result == 0:
            print("Clang lint passed")

        sys.exit(result)
    else:
        print("No files for Clang to process", file=sys.stderr)
        sys.exit(0)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

# Example: https://dmgbuild.readthedocs.io/en/latest/example.html

from __future__ import unicode_literals

import os.path

app = defines.get("app")
app_basename = os.path.basename(app)
format = defines.get("format", "UDBZ")
size = defines.get("size", None)
files = [app]
symlinks = {"Applications": "/Applications"}
icon = os.path.join(app, "Contents/Resources/Volume.icns")
icon_locations = {
    app_basename: (144, 190),
    "Applications": (455, 190),
}
background = os.path.join(app, "Contents/Resources/Background.tiff")
show_status_bar = False
show_tab_view = False
show_toolbar = False
show_pathbar = False
show_sidebar = False
sidebar_width = 180
window_rect = ((200, 120), (620, 420))
default_view = "icon-view"
show_icon_preview = False
include_icon_view_settings = "auto"
include_list_view_settings = "auto"
arrange_by = None
grid_offset = (0, 0)
grid_spacing = 100
scroll_position = (0, 0)
label_pos = "bottom"
text_size = 16
icon_size = 100
list_icon_size = 16
list_text_size = 12
list_scroll_position = (0, 0)
list_sort_by = "name"
list_use_relative_dates = True
list_calculate_all_sizes = (False,)
list_columns = ("name", "date-modified", "size", "kind", "date-added")
list_column_widths = {
    "name": 300,
    "date-modified": 181,
    "date-created": 181,
    "date-added": 181,
    "date-last-opened": 181,
    "size": 97,
    "kind": 115,
    "label": 100,
    "version": 75,
    "comments": 300,
}
list_column_sort_directions = {
    "name": "ascending",
    "date-modified": "descending",
    "date-created": "descending",
    "date-added": "descending",
    "date-last-opened": "descending",
    "size": "descending",
    "kind": "ascending",
    "label": "ascending",
    "version": "ascending",
    "comments": "ascending",
}



#!/usr/bin/env python3

# Deskflow -- mouse and keyboard sharing utility
# Copyright (C) 2024 Symless Ltd.
#
# This package is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# found in the file LICENSE that should have accompanied this file.
#
# This package is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import lib.env as env

env.ensure_in_venv(__file__)

import argparse, os, sys
import lib.cmd_utils as cmd_utils
import lib.colors as colors

valgrind_bin = "valgrind"


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--unit-tests", action="store_true")
    parser.add_argument("--integ-tests", action="store_true")
    parser.add_argument(
        "--filter-file",
        type=str,
        help="Takes the base filename without extension and uses it as a filter",
    )
    parser.add_argument(
        "--ignore-return-code",
        action="store_true",
        help="Ignore the return code of the test command",
    )
    parser.add_argument(
        "--valgrind",
        action="store_true",
        help="Run the test command with valgrind",
    )

    args = parser.parse_args()
    binary = get_binary_path(args)

    if args.filter_file:
        file_base = os.path.basename(args.filter_file)
        without_ext = os.path.splitext(file_base)[0]
        command = [binary, f"--gtest_filter={without_ext}*"]
    else:
        command = [binary]

    if args.valgrind:
        if not cmd_utils.has_command(valgrind_bin):
            print(f"{colors.ERROR_TEXT} {valgrind_bin} not found")
            sys.exit(1)

        command = [valgrind_bin] + command

    result = cmd_utils.run(command, print_cmd=True, check=False)
    if not args.ignore_return_code:
        sys.exit(result.returncode)


def get_binary_path(args):
    base_dir = "./build/bin"
    if args.unit_tests:
        return f"{base_dir}/unittests"
    elif args.integ_tests:
        return f"{base_dir}/integtests"
    else:
        raise RuntimeError("No test type specified")


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

# Deskflow -- mouse and keyboard sharing utility
# Copyright (C) 2024 Symless Ltd.
#
# This package is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# found in the file LICENSE that should have accompanied this file.
#
# This package is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import lib.env as env

env.ensure_in_venv(__file__, create_venv=True)
env.install_requirements()

import lib.colors as colors

print(colors.SUCCESS_TEXT, "Python virtual environment is ready.")



# Deskflow -- mouse and keyboard sharing utility
# Copyright (C) 2024 Symless Ltd.
#
# This package is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# found in the file LICENSE that should have accompanied this file.
#
# This package is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os, fnmatch


def find_files(search_dirs, include_files, exclude_dirs=[]):
    """Recursively find files, excluding specified directories"""
    matches = []
    for dir in search_dirs:
        for root, dirnames, filenames in os.walk(dir):
            dirnames[:] = [d for d in dirnames if d not in exclude_dirs]

            for pattern in include_files:
                for filename in fnmatch.filter(filenames, pattern):
                    matches.append(os.path.join(root, filename))
    return matches



import os, base64

# Deskflow -- mouse and keyboard sharing utility
# Copyright (C) 2024 Symless Ltd.
#
# This package is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# found in the file LICENSE that should have accompanied this file.
#
# This package is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

temp_path = "tmp/certificate"


class Certificate:
    """
    Installs a certificate from a base64 string, and returns the path to the certificate.
    Once the context is exited, the certificate is removed from the filesystem.

    Example usage:
    with Certificate(base64) as cert_path:
        print(f"Certificate path: {cert_path}")
    """

    def __init__(self, base64, file_ext):
        self.base64 = base64
        self.temp_filename = f"{temp_path}.{file_ext}"

    def __enter__(self):
        print(f"Decoding certificate to temporary path: {self.temp_filename}")
        try:
            cert_bytes = base64.b64decode(self.base64)
        except Exception as e:
            raise ValueError("Failed to decode certificate base64") from e

        os.makedirs(os.path.dirname(self.temp_filename), exist_ok=True)
        with open(self.temp_filename, "wb") as cert_file:
            cert_file.write(cert_bytes)

        return self.temp_filename

    def __exit__(self, _exc_type, _exc_value, _traceback):
        # not strictly necessary for ci, but when run on a dev machine, it reduces the risk
        # that private keys are left on the filesystem
        print(f"Removing temporary certificate file: {self.temp_filename}")
        os.remove(self.temp_filename)

        # propagate exceptions
        return False



#!/usr/bin/env python3

# Deskflow -- mouse and keyboard sharing utility
# Copyright (C) 2024 Symless Ltd.
#
# This package is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# found in the file LICENSE that should have accompanied this file.
#
# This package is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import lib.env as env

env.ensure_in_venv(__file__)

import argparse
import platform
from lib.linux import PackageType
from dotenv import load_dotenv  # type: ignore

ENV_FILE = ".env"
DEFAULT_PRODUCT_NAME = "Deskflow"
DEFAULT_FILENAME_BASE = "deskflow"
DEFAULT_PROJECT_BUILD_DIR = "build"
DEFAULT_DIST_DIR = "dist"
DEFAULT_TEST_CMD = "deskflow-server --version"
DEFAULT_PACKAGE_NAME = "deskflow"
VERSION_FILE = "VERSION"


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--leave-test-installed",
        action="store_true",
        help="Leave test package installed",
    )
    args = parser.parse_args()

    load_dotenv(dotenv_path=ENV_FILE)

    package(
        DEFAULT_FILENAME_BASE,
        get_app_version(VERSION_FILE),
        DEFAULT_PROJECT_BUILD_DIR,
        DEFAULT_DIST_DIR,
        DEFAULT_TEST_CMD,
        DEFAULT_PRODUCT_NAME,
        DEFAULT_PACKAGE_NAME,
        leave_test_installed=args.leave_test_installed,
    )


def get_app_version(filename):
    """
    Returns the version either from the env var, or from the version file.
    """
    version = env.get_env("DESKFLOW_VERSION", required=False)
    if version:
        return version

    with open(filename, "r") as f:
        return f.read().strip()


def package(
    filename_prefix,
    version,
    project_build_dir,
    dist_dir,
    test_cmd,
    product_name,
    package_name,
    source_dir=None,
    leave_test_installed=False,
):
    filename_base = get_filename_base(version, filename_prefix)
    print(f"Package filename base: {filename_base}")

    if env.is_windows():
        windows_package(filename_base, project_build_dir, dist_dir)
    elif env.is_mac():
        mac_package(
            filename_base, source_dir, project_build_dir, dist_dir, product_name
        )
    elif env.is_linux():
        linux_package(
            filename_base,
            filename_prefix,
            dist_dir,
            test_cmd,
            package_name,
            version,
            leave_test_installed,
        )
    else:
        raise RuntimeError(f"Unsupported platform: {env.get_os()}")


def get_filename_base(version, prefix, use_linux_distro=True):
    os = env.get_os()
    machine = platform.machine().lower()
    os_part = os

    if os == "linux" and use_linux_distro:
        distro_name, _distro_like, distro_version = env.get_linux_distro()
        if not distro_name:
            raise RuntimeError("Failed to detect Linux distro")

        if distro_version:
            version_for_filename = distro_version.replace(".", "-")
            os_part = f"{distro_name}-{version_for_filename}"
        else:
            os_part = distro_name

        # For consistency with existing filenames, we'll use 'amd64' instead of 'x86_64'.
        # Also, that's what Linux distros tend to call that architecture anyway.
        if machine == "x86_64":
            machine = "amd64"
    else:
        # Some Windows users get confused by 'amd64' and think it's 'arm64',
        # so we'll use Intel's 'x64' branding (even though it's wrong).
        # Also replace 'x86_64' with 'x64' for consistency.
        if machine == "amd64" or machine == "x86_64":
            machine = "x64"

    # Underscore is used to delimit different parts of the filename (e.g. version, OS, etc).
    # Dashes are used to delimit spaces, e.g. "debian-trixie" for "Debian Trixie".
    return f"{prefix}_{version}_{os_part}_{machine}"


def windows_package(filename_base, project_build_dir, dist_dir):
    import lib.windows as windows

    windows.package(filename_base, project_build_dir, dist_dir)


def mac_package(filename_base, source_dir, project_build_dir, dist_dir, product_name):
    import lib.mac as mac

    mac.package(filename_base, source_dir, project_build_dir, dist_dir, product_name)


def linux_package(
    filename_base,
    filename_prefix,
    dist_dir,
    test_cmd,
    package_name,
    version,
    leave_test_installed,
):
    import lib.linux as linux

    extra_packages = env.get_env_bool("LINUX_EXTRA_PACKAGES", False)

    linux.package(
        filename_base,
        dist_dir,
        test_cmd,
        package_name,
        PackageType.DISTRO,
        leave_test_installed,
    )

    if extra_packages:
        filename_base = get_filename_base(
            version, filename_base, use_linux_distro=False
        )
        linux.package(
            filename_prefix,
            dist_dir,
            test_cmd,
            package_name,
            PackageType.TGZ,
        )


if __name__ == "__main__":
    main()



# Deskflow -- mouse and keyboard sharing utility
# Copyright (C) 2024 Symless Ltd.
#
# This package is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# found in the file LICENSE that should have accompanied this file.
#
# This package is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os, shutil, glob, sys
import lib.cmd_utils as cmd_utils
import lib.env as env
from enum import Enum, auto

BUILD_ROOT_DIR = "build"


class PackageType(Enum):
    DISTRO = auto()
    TGZ = auto()


def run_command(command, check=True):

    has_sudo = cmd_utils.has_command("sudo")
    if "sudo" in command and not has_sudo:
        # assume we're running as root if sudo is not found (common on older distros).
        # a space char is intentionally added after "sudo" for intentionality.
        # possible limitation with stripping "sudo" is that if any packages with "sudo" in the
        # name are added to the list (probably very unlikely), this will have undefined behavior.
        print("The 'sudo' command was not found, stripping sudo from command")
        command = command.replace("sudo ", "").strip()

    cmd_utils.run(command, check, shell=True, print_cmd=True)


def package(
    filename_base,
    dist_dir,
    test_cmd,
    package_name,
    package_type: PackageType,
    leave_test_installed=False,
):
    working_dir = BUILD_ROOT_DIR
    extension, cmd = get_package_build_info(
        package_type,
    )
    run_package_cmd(cmd, working_dir)

    package_filename = get_package_filename(extension, working_dir)
    target_file = f"{filename_base}.{extension}"
    target_path = copy_to_dist_dir(package_filename, dist_dir, target_file)

    if package_type == PackageType.DISTRO:
        test_install(
            target_path, package_name, test_cmd, remove_test=not leave_test_installed
        )


def get_package_build_info(package_type: PackageType):

    command = None
    cpack_generator = None
    file_extension = None

    if package_type == PackageType.TGZ:
        cpack_generator = "TGZ"
        file_extension = "tar.gz"

    elif package_type == PackageType.DISTRO:

        distro, distro_like, _distro_version = env.get_linux_distro()
        if not distro_like:
            distro_like = distro

        if "debian" in distro_like:
            cpack_generator = "DEB"
            file_extension = "deb"
        elif "fedora" in distro_like or "opensuse" in distro_like:
            cpack_generator = "RPM"
            file_extension = "rpm"
        elif "arch" in distro_like:
            command = ["makepkg", "-s"]
            file_extension = "pkg.tar.zst"
        else:
            raise RuntimeError(f"Linux distro not yet supported: {distro}")

    if not cpack_generator and not command:
        raise RuntimeError("No package generator or command found")

    if cpack_generator:
        command = ["cpack", "-G", cpack_generator]

    return file_extension, command


def run_package_cmd(command, working_dir):
    package_user = env.get_env("LINUX_PACKAGE_USER", required=False)
    if package_user:
        cmd_utils.run(
            ["sudo", "chown", "-R", package_user, working_dir],
            check=True,
            print_cmd=True,
        )
        command = ["sudo", "-u", package_user] + command

    cwd = os.getcwd()
    try:
        os.chdir(working_dir)
        cmd_utils.run(command, check=True, print_cmd=True)
    finally:
        os.chdir(cwd)


def get_package_filename(extension, working_dir):
    files = glob.glob(f"{working_dir}/*.{extension}")

    if not files:
        raise ValueError(
            f"No .{extension} file found in build directory: {working_dir}"
        )

    return files[0]


def copy_to_dist_dir(source_file, dist_dir, target_file):
    os.makedirs(dist_dir, exist_ok=True)

    target_path = f"{dist_dir}/{target_file}"
    print(f"Copying to: {target_path}")
    shutil.copy(source_file, target_path)

    return target_path


def test_install(package_file, package_name, test_cmd, remove_test=True):

    distro, distro_like, _distro_version = env.get_linux_distro()
    if not distro_like:
        distro_like = distro

    install_base = None
    list_cmd = None
    remove_base = None
    if "debian" in distro_like:
        install_base = ["apt", "install", "-f", "-y"]
        remove_base = ["apt", "remove", "-y"]
        list_cmd = ["dpkg", "-L", package_name]
    elif "fedora" in distro_like:
        install_base = ["dnf", "install", "-y"]
        remove_base = ["dnf", "remove", "-y"]
        list_cmd = ["rpm", "-ql", package_name]
    elif "opensuse" in distro_like:
        install_base = ["zypper", "--no-gpg-checks", "install", "-y"]
        remove_base = ["zypper", "remove", "-y"]
        list_cmd = ["rpm", "-ql", package_name]
    elif "arch" in distro_like:
        install_base = ["pacman", "-U", "--noconfirm"]
        remove_base = ["pacman", "-R", "--noconfirm"]
        list_cmd = ["pacman", "-Ql", package_name]
    else:
        raise RuntimeError(f"Linux distro not yet supported: {distro}")

    has_sudo = cmd_utils.has_command("sudo")
    sudo = ["sudo"] if has_sudo else []

    print("Testing installation...")
    cmd_utils.run(
        sudo + install_base + [f"./{package_file}"],
        check=True,
        print_cmd=True,
    )

    print("Listing installed files...")
    cmd_utils.run(sudo + list_cmd, check=True, print_cmd=True)

    try:
        cmd_utils.run(test_cmd, shell=True, check=True, print_cmd=True)
    except Exception:
        raise RuntimeError("Unable to verify version")
    finally:
        if remove_test:
            cmd_utils.run(
                sudo + remove_base + [package_name], check=True, print_cmd=True
            )
        else:
            print("Leaving test package installed")

        print("Installation test passed")


def is_package_available(package):
    distro, distro_like, _distro_version = env.get_linux_distro()
    if not distro_like:
        distro_like = distro

    if "debian" in distro_like:
        command = ["apt-cache", "show", package]
    elif "fedora" in distro_like:
        command = ["dnf", "info", package]
    elif "opensuse" in distro_like:
        command = ["zypper", "info", package]
    elif "arch" in distro_like:
        command = ["pacman", "-Si", package]
    else:
        raise RuntimeError(f"Linux distro not yet supported: {distro}")

    result = cmd_utils.run(command, check=False, print_cmd=True, get_output=True)
    if result.stderr:
        print(result.stderr, file=sys.stderr)
    return result.returncode == 0



# Deskflow -- mouse and keyboard sharing utility
# Copyright (C) 2024 Symless Ltd.
#
# This package is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# found in the file LICENSE that should have accompanied this file.
#
# This package is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import git  # type: ignore
import lib.env as env
import lib.cmd_utils as cmd_utils

GIT_REPO = "https://github.com/microsoft/vcpkg.git"


def install(ci_env):
    vcpkg_bin = ensure_vcpkg(ci_env)

    cmd_utils.run([vcpkg_bin, "install"], print_cmd=True)


def ensure_vcpkg(ci_env):
    # Don't use the local vcpkg if we're in CI, since this makes caching complicated.
    if not ci_env and cmd_utils.has_command("vcpkg"):
        print("Using system vcpkg")
        return "vcpkg"

    if not os.path.exists("vcpkg"):
        get_vcpkg()
    else:
        print("Updating vcpkg...")
        repo = git.Repo("vcpkg")
        repo.remotes.origin.pull()

    if env.is_windows():
        vcpkg_bin = "vcpkg/vcpkg.exe"
    else:
        vcpkg_bin = "vcpkg/vcpkg"

    if not os.path.exists(vcpkg_bin):
        raise RuntimeError(f"Path not found: {vcpkg_bin}")

    return vcpkg_bin


def get_vcpkg():
    print("Downloading vcpkg...")
    git.Repo.clone_from(GIT_REPO, "vcpkg")

    os.chdir("vcpkg")
    try:
        if env.is_windows():
            cmd_utils.run("bootstrap-vcpkg.bat", shell=True, print_cmd=True)
        else:
            cmd_utils.run("./bootstrap-vcpkg.sh", shell=True, print_cmd=True)
    finally:
        os.chdir("..")



# Deskflow -- mouse and keyboard sharing utility
# Copyright (C) 2024 Symless Ltd.
#
# This package is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# found in the file LICENSE that should have accompanied this file.
#
# This package is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os, sys
import lib.cmd_utils as cmd_utils
import lib.env as env

build_dir = "build/meson"


def meson_venv_bin():
    if not env.in_venv():
        raise RuntimeError("Not in a virtual environment")

    return os.path.join(os.path.dirname(sys.executable), "meson")


def setup(no_system_list, static_list):
    cmd = [meson_venv_bin(), "setup", build_dir]

    # TODO: These special Windows exceptions should probably be in Meson
    # or somewhere other than this script, as it's a bit hacky.
    if env.is_windows():
        cmd.append("-Dsystem-gtest=false")
        cmd.append("-Dsystem-tomlplusplus=false")
        cmd.append("-Dsystem-cli11=false")

    for subproject in no_system_list or []:
        cmd.append(f"-Dsystem-{subproject}=false")

    # This might be a bit rude, but Meson seems to cache a lot (like CMake),
    # so wiping every time is the easiest way to ensure that the build is clean.
    # Plus, the way we're using Meson (at the moment) is just for satisfying
    # dependencies, so this script is run infrequently enough to not matter.
    if os.path.exists(build_dir):
        cmd.append("--wipe")

    cmd_utils.run(cmd, print_cmd=True)

    for subproject in static_list or []:
        static_subproject(subproject)


def static_subproject(subproject):
    if subproject == "libportal":
        # HACK: This is a bit horrible. Ideally, Meson would take care of applying this patch,
        # but it only seems to copy the .diff over and not apply the patch.
        #
        # Important: Static linking is not intended for package maintainers, only for beta testers.
        # Static linking is also pretty horrible, but many distros will be slow to pick up 0.8.x
        # which has input capture support. The sooner we can remove this patching code the better.
        cmd_utils.run(
            [
                "patch",
                "-d",
                "subprojects/libportal",
                "-p1",
                "-i",
                "static-lib.diff",
            ],
            print_cmd=True,
            check=False,
        )
    else:
        raise RuntimeError(f"Unknown subproject: {subproject}")


def compile():
    cmd_utils.run([meson_venv_bin(), "compile", "-C", build_dir], print_cmd=True)


def install():
    cmd = [meson_venv_bin(), "install", "-C", build_dir]

    has_sudo = cmd_utils.has_command("sudo")
    if has_sudo:
        cmd.insert(0, "sudo")

    cmd_utils.run(cmd, print_cmd=True)



# Deskflow -- mouse and keyboard sharing utility
# Copyright (C) 2024 Symless Ltd.
#
# This package is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# found in the file LICENSE that should have accompanied this file.
#
# This package is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os

github_env_key = "GITHUB_ENV"
github_path_key = "GITHUB_PATH"
github_output_key = "GITHUB_OUTPUT"


def set_env_var(key, value):
    """
    Appends the key=value pair to the GitHub environment file.
    """
    env_file = os.getenv(github_env_key)
    if not env_file:
        raise RuntimeError(f"Env var {github_env_key} not set")

    if not os.path.exists(env_file):
        raise RuntimeError(f"Env file not found: {env_file}")

    print(f"Setting GitHub env var: {key}={value}")
    with open(env_file, "a") as env_file:
        env_file.write(f"{key}={value}\n")


def add_to_path(value):
    """
    Appends the value to the GitHub path.
    """
    path_file = os.getenv(github_path_key)
    if not path_file:
        raise RuntimeError(f"Env var {github_path_key} not set")

    if not os.path.exists(path_file):
        raise RuntimeError(f"Path file not found: {path_file}")

    print(f"Adding to GitHub path: {value}")
    with open(path_file, "a") as path_file:
        path_file.write(f"{value}\n")


def set_output(name, value):
    """
    Sets the output parameter for the GitHub action.
    """
    output_file = os.getenv(github_output_key)
    if not output_file:
        raise RuntimeError(f"Env var {github_output_key} not set")

    if not os.path.exists(output_file):
        raise RuntimeError(f"Output file not found: {output_file}")

    print(f"Setting GitHub output: {name}={value}")
    with open(output_file, "a") as output_file:
        output_file.write(f"{name}={value}\n")



# Deskflow -- mouse and keyboard sharing utility
# Copyright (C) 2024 Symless Ltd.
#
# This package is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# found in the file LICENSE that should have accompanied this file.
#
# This package is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os, sys, subprocess
import lib.cmd_utils as cmd_utils

# The `.venv` dir seems to be most common for virtual environments.
VENV_DIR = ".venv"


def check_module(module):
    try:
        __import__(module)
        return True
    except ImportError:
        print(f"Python is missing {module} module", file=sys.stderr)
        return False


def get_os():
    """Detects the operating system."""
    if sys.platform == "win32":
        return "windows"
    elif sys.platform == "darwin":
        return "mac"
    elif sys.platform.startswith("linux"):
        return "linux"
    else:
        raise RuntimeError(f"Unsupported platform: {sys.platform}")


def is_windows():
    return get_os() == "windows"


def is_mac():
    return get_os() == "mac"


def is_linux():
    return get_os() == "linux"


def get_linux_distro():
    """Detects the Linux distro."""
    os_file = "/etc/os-release"
    name = None
    name_like = None
    version_id = None
    version_codename = None

    if os.path.isfile(os_file):
        with open(os_file) as f:
            for line in f:
                if line.startswith("ID="):
                    name = line.strip().split("=")[1].strip('"')
                elif line.startswith("ID_LIKE="):
                    name_like = line.strip().split("=")[1].strip('"')
                elif line.startswith("VERSION_ID="):
                    version_id = line.strip().split("=")[1].strip('"')
                elif line.startswith("VERSION_CODENAME="):
                    version_codename = line.strip().split("=")[1].strip('"')

    return name, name_like, version_id or version_codename


def get_env(name, required=True, default=None):
    """
    Returns an env var (stripped) or optionally raises an error if not set.

    If `default` is set, it will be returned even if `required` is True.
    """
    value = os.getenv(name)
    if value:
        value = value.strip()

    if not value:
        if default:
            return default
        elif required:
            raise ValueError(f"Required env var not set: {name}")

    return value


def get_env_bool(name, default=False):
    """Returns a boolean value from an env var (stripped)."""
    value = os.getenv(name)
    if value:
        value = value.strip()

    if value is None:
        return default

    return value.lower() in ["true", "1", "yes"]


def get_venv_executable(binary="python"):
    if sys.platform == "win32":
        return os.path.join(VENV_DIR, "Scripts", binary)
    else:
        return os.path.join(VENV_DIR, "bin", binary)


def in_venv():
    """Returns True if the script is running in a Python virtual environment."""
    return sys.prefix != sys.base_prefix


def ensure_in_venv(script_file, create_venv=False):
    """
    Ensures the script is running in a Python virtual environment (venv).
    If the script is not running in a venv, it will create one and re-run the script in the venv.
    """

    check_dependencies(raise_error=True)
    import venv

    if in_venv():
        print(f"Running in venv, executable: {sys.executable}", flush=True)
        return

    if create_venv and not os.path.exists(VENV_DIR):
        print(f"Creating virtual environment at {VENV_DIR}")
        venv.create(VENV_DIR, with_pip=True)

    if os.path.exists(VENV_DIR):
        script_file_abs = os.path.abspath(script_file)
        print(f"Using virtual environment for: {script_file_abs}", flush=True)
        python_executable = get_venv_executable()
        result = subprocess.run([python_executable, script_file_abs] + sys.argv[1:])
        sys.exit(result.returncode)
    else:
        print(
            "The Python virtual environment (.venv) needs to be created before you can "
            "run this script.\n"
            "Please run: scripts/setup_venv.py"
        )
        sys.exit(1)


def install_requirements():
    """
    Uses `pip` to install required Python modules from the `requirements.txt` file.
    """

    check_dependencies(raise_error=True)

    print("Updating pip...")
    cmd_utils.run(
        [sys.executable, "-m", "pip", "install", "--upgrade", "pip"],
        shell=False,
        print_cmd=True,
    )

    print("Installing required modules...")
    cmd_utils.run(
        [sys.executable, "-m", "pip", "install", "-e", "scripts"],
        shell=False,
        print_cmd=True,
    )


def check_dependencies(raise_error=False):
    """
    Returns True if pip and venv are available.
    """

    has_pip = check_module("pip")
    has_venv = check_module("venv")

    if raise_error:
        if not has_pip:
            raise RuntimeError("Python is missing pip")
        if not has_venv:
            raise RuntimeError("Python is missing venv")
    else:
        return has_pip and has_venv


def ensure_dependencies():
    """
    Ensures that pip and venv are available, and installs them if they are not.
    This is normally only installs on Linux, as Windows and Mac usually come with pip and venv.
    """

    if check_dependencies():
        return

    print("Installing Python dependencies...")

    os = get_os()
    if os != "linux":
        # should not be a problem, since windows and mac come with pip and venv
        raise RuntimeError(f"Unable to install Python dependencies on {os}")

    has_sudo = cmd_utils.has_command("sudo")
    sudo = "sudo" if has_sudo else ""

    distro, distro_like, _version = get_linux_distro()
    if not distro_like:
        distro_like = distro

    update_cmd = None
    install_cmd = None
    if distro == "rhel" or "rhel" in distro_like:
        update_cmd = "yum check-update"
        install_cmd = "yum install -y python3-pip"  # rhel-like has venv already
    elif "debian" in distro_like:
        update_cmd = "apt update"
        install_cmd = "apt install -y python3-pip python3-venv"
    elif "fedora" in distro_like:
        update_cmd = "dnf check-update"
        install_cmd = "dnf install -y python3-pip python3-virtualenv"
    elif "arch" in distro_like:
        install_cmd = "pacman -Syu --noconfirm python-pip python-virtualenv"
    elif "opensuse" in distro_like:
        update_cmd = "zypper refresh"
        install_cmd = "zypper install -y python3-pip python3-virtualenv"
    else:
        raise RuntimeError(f"Unable to install Python dependencies on {distro}")

    if update_cmd:
        # don't check the return code, as some package managers return non-zero exit codes
        # under normal circumstances (e.g. dnf check-update returns 100 when there are
        # updates available).
        cmd_utils.run(
            f"{sudo} {update_cmd}".strip(), check=False, shell=True, print_cmd=True
        )

    cmd_utils.run(f"{sudo} {install_cmd}".strip(), shell=True, print_cmd=True)


def import_colors():
    import lib.colors as colors

    return colors


def persist_lock_file(path):
    """
    Persists a lock file and ensures the directory part of the path exists.
    """
    dir_path = os.path.dirname(path)
    if not os.path.exists(dir_path):
        os.makedirs(dir_path, exist_ok=True)

    with open(path, "w") as f:
        f.write(str(os.getpid()))


def remove_lock_file(path):
    """
    Removes a lock file if it exists.
    """
    if os.path.exists(path):
        os.remove(path)



# Deskflow -- mouse and keyboard sharing utility
# Copyright (C) 2024 Symless Ltd.
#
# This package is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# found in the file LICENSE that should have accompanied this file.
#
# This package is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os, sys
import lib.cmd_utils as cmd_utils
import glob


class Qt:
    def __init__(
        self, mirror_url, version, base_dir, modules, os_name, compiler, tool_dir
    ):
        self.mirror_url = mirror_url
        self.version = version
        self.base_dir = base_dir
        self.modules = modules
        self.os_name = os_name
        self.compiler = compiler
        self.tool_dir = tool_dir
        self.dir_pattern = f"{self.base_dir}{os.sep}{self.version}*/{self.tool_dir}"

    def get_install_dir(self):
        match = glob.glob(self.dir_pattern)
        return os.path.abspath(match[0]) if match else None

    def install(self):
        """Install Qt."""

        if self.get_install_dir():
            print(f"Skipping Qt, already installed at: {self.dir_pattern}")
            return

        args = [sys.executable, "-m", "aqt", "install-qt"]
        args.extend(["--outputdir", self.base_dir])
        args.extend(["--base", self.mirror_url])
        args.extend([self.os_name, "desktop", str(self.version), self.compiler])

        if self.modules:
            args.extend(["-m"] + self.modules)

        print(args)
        cmd_utils.run(
            args,
            print_cmd=True,
        )

        if not self.get_install_dir():
            raise RuntimeError(
                f"Qt was not installed, path not found: {self.dir_pattern}"
            )


class WindowsQt(Qt):
    def __init__(self, mirror_url, version, base_dir, modules):
        super().__init__(
            mirror_url,
            version,
            base_dir,
            modules,
            "windows",
            "win64_msvc2019_64",
            "msvc2019_64",
        )


class MacQt(Qt):
    def __init__(self, mirror_url, version, base_dir, modules):
        super().__init__(
            mirror_url,
            version,
            base_dir,
            modules,
            "mac",
            "clang_64",
            "macos",
        )



# Deskflow -- mouse and keyboard sharing utility
# Copyright (C) 2024 Symless Ltd.
#
# This package is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# found in the file LICENSE that should have accompanied this file.
#
# This package is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import glob, os, shutil, sys
import lib.env as env
import lib.colors as colors


class CopyOptions:
    def __init__(self, ignore_errors, verbose):
        self.ignore_errors = ignore_errors
        self.verbose = verbose


class CopyContext:
    def __init__(self):
        self.errors = 0
        self.permission_error = False


def copy(source, target, options):
    """Copy files and directories from source to target."""

    context = CopyContext()
    if options.verbose:
        print(f"Copying files from {source} to {target}")

    try:
        for match in glob.glob(source):

            if os.path.isfile(match):
                copy_file(match, target, options, context)
            elif os.path.isdir(match):
                copy_dir(match, target, options, context)
            else:
                raise RuntimeError(f"Path {match} is not a file or directory")
    finally:
        if context.errors and options.ignore_errors:
            print(f"{colors.WARNING_TEXT} Ignored {context.errors} copy error(s)")

        if context.permission_error and env.is_windows():
            print(
                f"{colors.HINT_TEXT} A Windows file permission error may mean that the file is in use"
            )


def copy_dir(match, target, options, context):
    if options.verbose:
        print(f"Copying directory {match} to {target}")

    try:
        shutil.copytree(match, target, dirs_exist_ok=True)
    except Exception as e:
        handle_all_copy_errors(e, options, context)


def copy_file(match, target, options, context):
    if options.verbose:
        print(f"Copying file {match} to {target}")

    try:
        shutil.copy(match, target)
    except Exception as e:
        handle_all_copy_errors(e, options, context)


def handle_all_copy_errors(error, options, context):

    if isinstance(error, shutil.Error):
        for _, _, file_error in error.args[0]:
            handle_copy_error(file_error, options, context)
    else:
        handle_copy_error(error, options, context)

    if not options.ignore_errors:
        raise error


def handle_copy_error(error, options, context):
    if isinstance(error, PermissionError):
        context.permission_error = True

    if isinstance(error, str):
        context.permission_error = error.startswith("[Errno 13] Permission denied")

    context.errors += 1

    if options.ignore_errors:
        print(
            f"{colors.WARNING_TEXT} Copy failed: {error}",
            file=sys.stderr,
        )



#!/usr/bin/env python3

# Deskflow -- mouse and keyboard sharing utility
# Copyright (C) 2024 Symless Ltd.
#
# This package is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# found in the file LICENSE that should have accompanied this file.
#
# This package is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import lib.env as env

env.ensure_in_venv(__file__)

import os, sys, argparse
import lib.windows as windows
import lib.colors as colors

DEFAULT_SERVICE_ID = "deskflow"
DEFAULT_BIN_NAME = "deskflow-daemon"
DEFAULT_SOURCE_DIR = os.path.join("build", "temp", "bin")
DEFAULT_TARGET_DIR = os.path.join("build", "bin")
IGNORE_PROCESSES = ["deskflow.exe"]


def main():
    """Entry point for the script."""

    parser = argparse.ArgumentParser()
    parser.add_argument("--reinstall", action="store_true")
    parser.add_argument("--stop", action="store_true")
    parser.add_argument("--restart", action="store_true")
    parser.add_argument("--pause-on-exit", action="store_true")
    parser.add_argument("--source-dir", default=DEFAULT_SOURCE_DIR)
    parser.add_argument("--target-dir", default=DEFAULT_TARGET_DIR)
    parser.add_argument("--bin-name", default=DEFAULT_BIN_NAME)
    parser.add_argument("--ignore-processes", nargs="+", default=IGNORE_PROCESSES)
    parser.add_argument("--service-id", default=DEFAULT_SERVICE_ID)
    parser.add_argument("--verbose", action="store_true")
    args = parser.parse_args()

    if not env.is_windows():
        print(
            f"{colors.ERROR_TEXT} This script is only supported on Windows",
            file=sys.stderr,
        )
        sys.exit(1)

    service = windows.WindowsService(__file__, args)

    try:
        if args.reinstall:
            service.reinstall()
        elif args.stop:
            service.stop()
        elif args.restart:
            service.restart()
        else:
            print("No action specified", file=sys.stderr)
            exit(1)
    except Exception as e:
        print(f"{colors.ERROR_TEXT} {e}", file=sys.stderr)

    if args.pause_on_exit:
        input("Press enter to continue...")


if __name__ == "__main__":
    main()



# Deskflow -- mouse and keyboard sharing utility
# Copyright (C) 2024 Symless Ltd.
#
# This package is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# found in the file LICENSE that should have accompanied this file.
#
# This package is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import yaml
import lib.env as env
import lib.cmd_utils as cmd_utils

config_file = "config.yaml"
root_key = "config"
deps_key = "dependencies"
command_key = "command"
command_pre_key = "command-pre"
subprojects_key = "subprojects"
arrow = " ➤ "


class ConfigKeyError(RuntimeError):
    def __init__(self, config_file, key):
        self.config_file = config_file
        self.key = key

    def __str__(self):
        return f"Not found in {self.config_file}: {self.key}"


def _get(dict, key, key_parent=None, required=True):
    value = dict.get(key)

    if required and not value:
        key_path = f"{root_key}{arrow}{key_parent}{arrow}{key}" if key_parent else key
        raise ConfigKeyError(config_file, key_path)

    return value


class Config:
    """Reads the project configuration YAML file."""

    def __init__(self):

        with open(config_file, "r") as f:
            data = yaml.safe_load(f)

        self.os_name = env.get_os()

        print("Config for OS:", self.os_name)
        self.root = _get(data, root_key)
        self.os = _get(self.root, self.os_name)

    def get_os_value(self, key, required=True, linux_distro=None):
        if linux_distro:
            # recurse with the linux distro as the key parameter to get the base distro key.
            distro = self.get_os_value(key=linux_distro)
            return _get(distro, key, f"{self.os_name}{arrow}{linux_distro}", required)
        else:
            return _get(self.os, key, self.os_name, required)

    def get_qt_config(self):
        qt_key = "qt"
        qt = self.get_os_deps_value(qt_key)

        parent_key = f"{self.os_name}{arrow}{deps_key}"
        mirror_url = _get(qt, "mirror", parent_key)
        version = _get(qt, "version", parent_key)
        base_dir = _get(qt, "base-dir", parent_key)
        modules = _get(qt, "modules", parent_key, required=False)

        return mirror_url, version, base_dir, modules

    def get_os_deps_value(self, key, required=True, linux_distro=None):
        deps = self.get_os_value(deps_key, required, linux_distro)
        if linux_distro:
            key_parent = f"{self.os_name}{arrow}{linux_distro}{arrow}{deps_key}"
        else:
            key_parent = f"{self.os_name}{arrow}{deps_key}"
        return _get(deps, key, key_parent, required)

    def get_os_deps_command(self, key=command_key, required=True, linux_distro=None):
        command = self.get_os_deps_value(key, required, linux_distro)
        if command:
            return cmd_utils.strip_continuation_sequences(command)
        else:
            return None

    def get_os_subprojects(self):
        distro, _distro_like, _distro_version = env.get_linux_distro()
        return self.get_os_value(subprojects_key, linux_distro=distro, required=False)

    def get_subproject_deps_command(self, subproject_name):
        subprojects = _get(self.root, subprojects_key)
        subproject = _get(subprojects, subproject_name, subprojects_key)
        deps_parent = f"{subprojects_key}{arrow}{subproject_name}"
        deps = _get(subproject, deps_key, deps_parent)

        if env.is_linux():
            distro, _distro_like, _distro_version = env.get_linux_distro()
            if not distro:
                raise RuntimeError("Unable to detect Linux distro")

            command = _get(deps, distro, f"{deps_parent}{arrow}{deps_key}")
        else:
            command = _get(deps, self.os_name, f"{deps_parent}{arrow}{deps_key}")

        return cmd_utils.strip_continuation_sequences(command)

    def get_os_deps_command_pre(self, required=True, linux_distro=None):
        return self.get_os_deps_command(command_pre_key, required, linux_distro)

    def get_windows_ci_config(self):
        choco_ci_key = "ci"
        choco_ci = self.get_os_deps_value(choco_ci_key)

        choco_ci_path = f"{self.os_name}{arrow}{deps_key}{arrow}{choco_ci_key}"
        edit_config = _get(choco_ci, "edit-config", choco_ci_path)
        skip_packages = _get(choco_ci, "skip-packages", choco_ci_path)

        return edit_config, skip_packages



#!/usr/bin/env python3

# Deskflow -- mouse and keyboard sharing utility
# Copyright (C) 2024 Symless Ltd.
#
# This package is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# found in the file LICENSE that should have accompanied this file.
#
# This package is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import lib.env as env

env.ensure_in_venv(__file__)

import argparse
import lib.file_utils as file_utils
import lib.colors as colors


def main():
    """
    Cross platform script to copy files and directories.
    This script was mostly created beause the default `copy` command on Windows is too noisy.
    If this becomes complex it must be replaced with a library.
    """

    parser = argparse.ArgumentParser()
    parser.add_argument("source", help="Source pattern to copy from")
    parser.add_argument("target", help="Destination pattern to copy to")
    parser.add_argument(
        "--ignore-errors", action="store_true", help="Ignore errors when copying"
    )
    parser.add_argument(
        "--verbose", action="store_true", help="Print more information to the console"
    )
    args = parser.parse_args()

    options = file_utils.CopyOptions(args.ignore_errors, args.verbose)

    try:
        file_utils.copy(args.source, args.target, options)
    except Exception as e:
        if not args.ignore_errors:
            raise e
        else:
            print(f"{colors.ERROR_TEXT} {e}")


if __name__ == "__main__":
    main()



# Deskflow -- mouse and keyboard sharing utility
# Copyright (C) 2024 Symless Ltd.
#
# This package is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# found in the file LICENSE that should have accompanied this file.
#
# This package is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import colorama  # type: ignore
from colorama import Fore  # type: ignore

colorama.init()

SUCCESS_TEXT = f"{Fore.LIGHTGREEN_EX}Success:{Fore.RESET}"
ERROR_TEXT = f"{Fore.LIGHTRED_EX}Error:{Fore.RESET}"
WARNING_TEXT = f"{Fore.LIGHTYELLOW_EX}Warning:{Fore.RESET}"
HINT_TEXT = f"{Fore.LIGHTBLUE_EX}Hint:{Fore.RESET}"



#!/usr/bin/env python3

# Deskflow -- mouse and keyboard sharing utility
# Copyright (C) 2024 Symless Ltd.
#
# This package is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# found in the file LICENSE that should have accompanied this file.
#
# This package is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os, sys, argparse, traceback
import lib.env as env
import lib.cmd_utils as cmd_utils
import lib.qt_utils as qt_utils
import lib.github as github
import lib.meson as meson

path_env_var = "PATH"
cmake_prefix_env_var = "CMAKE_PREFIX_PATH"


def main():
    is_ci = os.getenv("CI") is not None
    if is_ci:
        print("CI environment detected")

    args = parse_args(is_ci)
    if args.lock_file:
        env.persist_lock_file(args.lock_file)

    try:
        run(args)
    except Exception:
        traceback.print_exc()
        sys.exit(1)
    finally:
        if env.is_windows() and args.pause_on_exit:
            # Allow the rest of the install to continue while sitting at the pause.
            if args.lock_file:
                env.remove_lock_file(args.lock_file)

            # Useful on Windows, when elevated, Python is opened in a new window and closes
            # immediately after the script finishes. This keeps the script window open so that
            # the user can see the output.
            print()
            input("Press enter to continue...")


def parse_args(is_ci):
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "--ci-env",
        action="store_true",
        help="Useful for faking CI env (defaults to true in CI env)",
        default=is_ci,
    )
    parser.add_argument(
        "--lock-file",
        type=str,
        help="Create a file to indicate script is running",
    )
    parser.add_argument(
        "--only-python", action="store_true", help="Only install Python dependencies"
    )
    parser.add_argument(
        "--skip-python",
        action="store_true",
        help="Do not install Python dependencies",
    )
    parser.add_argument(
        "--skip-system",
        action="store_true",
        help="Do not install system dependencies (apt, dnf, etc)",
    )
    parser.add_argument(
        "--skip-meson", action="store_true", help="Do not setup and compile with Meson"
    )
    parser.add_argument(
        "--subprojects",
        action="store_true",
        help="Install dependencies for Meson subprojects (use with --meson-no-system)",
    )
    parser.add_argument(
        "--meson-install",
        action="store_true",
        help="Install built Meson subprojects to system",
    )
    parser.add_argument(
        "--meson-no-system",
        nargs="+",
        help="Specify which Meson subprojects to use instead of system dependencies",
    )
    parser.add_argument(
        "--meson-static",
        nargs="+",
        help="Specify which Meson subprojects to build as static libraries",
    )

    if env.is_windows():
        parser.add_argument(
            "--skip-vcpkg",
            action="store_true",
            help="Windows only: Do not install vcpkg dependencies",
        )
        parser.add_argument(
            "--skip-elevated",
            action="store_true",
            help="Windows only: Do not run elevated command",
        )
        parser.add_argument(
            "--only-elevated",
            action="store_true",
            help="Windows only: Only run elevated command",
        )
        parser.add_argument(
            "--pause-on-exit",
            action="store_true",
            help="Windows only: Useful to prevent elevated window from closing",
        )

    return parser.parse_args()


def run(args):
    env.ensure_dependencies()
    env.ensure_in_venv(__file__, create_venv=True)

    if not args.skip_python:
        env.install_requirements()

    colors = env.import_colors()

    if args.only_python:
        print()
        print(colors.SUCCESS_TEXT + " Only Python dependencies installed")
        return

    try:
        install(args)

        print()
        print(f"\n{colors.SUCCESS_TEXT} Dependencies installed")

        # On Windows and macOS, we set env vars for cmake, but for them to be picked up,
        # either the shell needs to be restarted or the env vars need to be re-sourced.
        # Restarting the shell is easier for most people.
        if not env.is_linux():
            print(
                f"{colors.WARNING_TEXT} You may need to restart your terminal "
                "or IDE to use new env vars"
            )
    except Exception:
        traceback.print_exc()
        print()
        print(f"\n{colors.ERROR_TEXT} Failed to install dependencies")
        sys.exit(1)


def install(args):
    if not args.skip_system:
        deps = Dependencies(args)
        deps.install()

    # Only install vcpkg dependencies on Windows, since on other OS it's not needed (yet).
    # We probably won't ever need this on macOS and Linux since brew and apt/dnf/etc do a
    # good job of providing dependencies. Where they don't, we can use Meson.
    if env.is_windows() and not args.skip_vcpkg:
        import lib.vcpkg as vcpkg

        vcpkg.install(args.ci_env)

    if not args.skip_meson:
        if args.subprojects:
            for subproject in args.meson_no_system or []:
                deps = SubprojectDependencies(subproject)
                deps.install()

        run_meson(args.meson_install, args.meson_no_system, args.meson_static)


# It's a bit weird to use Meson just for installing deps, but it's a stopgap until
# we fully switch from CMake to Meson. For the meantime, Meson will install the deps
# so that CMake can find them easily. Once we switch to Meson, it might be possible for
# Meson handle the deps resolution, so that we won't need to install them on the system.
def run_meson(install, no_system_list, static_list):
    meson.setup(no_system_list, static_list)

    # Only compile and install on Linux for now, since we're only using Meson to fetch
    # the deps on Windows and macOS.
    if env.is_linux():
        meson.compile()

    if install:
        meson.install()


class Dependencies:

    def __init__(self, args):
        from lib.config import Config

        self.config = Config()
        self.args = args
        self.ci_env = args.ci_env

    def install(self):
        """Installs dependencies for the current platform."""

        if env.is_windows():
            self.windows()
        elif env.is_mac():
            self.mac()
        elif env.is_linux():
            self.linux()
        else:
            raise RuntimeError(f"Unsupported platform: {os}")

    def windows(self):
        """Installs dependencies on Windows."""
        import lib.windows as windows

        if not self.args.skip_elevated:
            if windows.is_admin():

                # The choco command should run from the elevated command.
                choco = windows.WindowsChoco()
                choco.ensure_choco_installed()
                command_elevated = self.config.get_os_deps_command("command-elevated")
                cmd_utils.run(command_elevated, shell=True, print_cmd=True)

                if self.args.only_elevated:
                    sys.exit(0)
            else:
                windows.run_elevated(
                    __file__, "--only-elevated --skip-python", wait_for_exit=True
                )

        qt = qt_utils.WindowsQt(*self.config.get_qt_config())
        qt.install()

        if self.ci_env:
            github.set_env_var(cmake_prefix_env_var, qt.get_install_dir())
        else:
            windows.set_env_var(cmake_prefix_env_var, qt.get_install_dir())

        command = self.config.get_os_deps_command()

        cmd_utils.run(command, shell=True, print_cmd=True)

    def mac(self):
        """Installs dependencies on macOS."""
        import lib.mac as mac

        # On macOS, brew does have a Qt package available, but it is always built against the
        # current macOS version and the brew version also does some really weird stuff with the
        # library symbols, which confuses the heck out of `macqtdeploy`. So, using the official
        # Qt library binaries seems to be the most reliable option for distribution.
        qt = qt_utils.MacQt(*self.config.get_qt_config())
        qt.install()

        qt_dir = qt.get_install_dir()
        qt_bin_dir = os.path.join(qt_dir, "bin")
        env_vars_set = 0
        if self.ci_env:
            github.set_env_var(cmake_prefix_env_var, qt_dir)
            github.add_to_path(qt_bin_dir)
        else:
            env_vars_set += mac.set_env_var(cmake_prefix_env_var, qt_dir)
            env_vars_set += mac.set_env_var(path_env_var, qt_bin_dir)

        command = self.config.get_os_deps_command()
        cmd_utils.run(command, shell=True, print_cmd=True)

        if env_vars_set:
            print(f"To load env vars, run: source {mac.SHELL_RC}")

    def linux(self):
        """Installs dependencies on Linux."""
        import lib.linux as linux

        distro, distro_like, _distro_version = env.get_linux_distro()
        if not distro:
            raise RuntimeError("Unable to detect Linux distro")

        command_pre = self.config.get_os_deps_command_pre(
            linux_distro=distro, required=False
        )
        if command_pre:
            print("Running dependencies prerequisites command")

            check = True
            if distro == "fedora" or (distro_like and "fedora" in distro_like):
                print(
                    "Fedora-like detected, "
                    "ignoring return code on dependencies prerequisites command"
                )
                # On Fedora-like, dnf update returns code 100 when updates are available.
                check = False

            linux.run_command(command_pre, check)

        command = self.config.get_os_deps_command(linux_distro=distro)
        optional = self.config.get_os_deps_value(
            "optional", linux_distro=distro, required=False
        )
        for optional_package in optional or []:
            if not linux.is_package_available(optional_package):
                print(f"Optional package not found, stripping: {optional_package}")
                command = command.replace(optional_package, "")

        linux.run_command(command, check=True)

        subprojects = self.config.get_os_subprojects()
        if subprojects:
            for subproject in subprojects:
                deps = SubprojectDependencies(subproject)
                deps.install()


class SubprojectDependencies:

    def __init__(self, subproject):
        from lib.config import Config

        self.subproject = subproject
        self.config = Config()

    def install(self):
        """Installs dependencies for the current platform."""

        print(f"Installing dependencies for sub-project: {self.subproject}")

        if env.is_linux():
            self.linux()
        else:
            raise RuntimeError(f"Unsupported platform: {os}")

    def linux(self):
        """Installs dependencies on Linux."""
        import lib.linux as linux

        command = self.config.get_subproject_deps_command(self.subproject)
        linux.run_command(command, check=True)


if __name__ == "__main__":
    main()



# Deskflow -- mouse and keyboard sharing utility
# Copyright (C) 2024 Symless Ltd.
#
# This package is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# found in the file LICENSE that should have accompanied this file.
#
# This package is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import dmgbuild  # type: ignore
import os, time, json, shutil, sys
import lib.cmd_utils as cmd_utils
import lib.env as env
from lib.certificate import Certificate

CERT_P12_ENV = "APPLE_P12_CERTIFICATE"
NOTARY_USER_ENV = "APPLE_NOTARY_USER"
CODESIGN_ENV = "APPLE_CODESIGN_ID"
SHELL_RC = "~/.zshrc"
SETTINGS_FILE = "res/dist/mac/dmgbuild/settings.py"
SECURITY_PATH = "/usr/bin/security"
SUDO_PATH = "/usr/bin/sudo"
NOTARYTOOL_PATH = "/usr/bin/notarytool"
CODESIGN_PATH = "/usr/bin/codesign"
XCODE_SELECT_PATH = "/usr/bin/xcode-select"
KEYCHAIN_PATH = "/Library/Keychains/System.keychain"


def set_env_var(name, value):
    """
    Adds to an environment variable in the shell rc file.

    Returns True if the variable was added, False if it already exists.
    """
    text = f'export {name}="{value}:${name}"'
    file = os.path.expanduser(SHELL_RC)
    if os.path.exists(file):
        with open(file, "r") as f:
            if text in f.read():
                return False

    print(f"Setting environment variable: {name}={name}")
    with open(file, "a") as f:
        f.write(f"\n{text}\n")
        print(f"Appended to {SHELL_RC}: {text}")
        return True


def package(filename_base, source_dir, build_dir, dist_dir, product_name):
    """
    Package the application for macOS.
    The app bundle must be signed, or an error will occur:
    > EXC_BAD_ACCESS (SIGKILL (Code Signature Invalid))
    An "Apple Development" certificate is sufficient for local development.
    """

    (
        codesign_id,
        cert_base64,
        cert_password,
        notary_user,
        notary_password,
        notary_team_id,
    ) = package_env_vars()

    if cert_base64:
        install_certificate(cert_base64, cert_password)
    else:
        print(
            f"Warning: Skipped certificate installation, env var {CERT_P12_ENV} not set",
            file=sys.stderr,
        )

    bundle_source_dir = os.path.join(
        build_dir, os.path.join("bundle", product_name + ".app")
    )
    build_bundle(bundle_source_dir)

    if codesign_id:
        sign_bundle(bundle_source_dir, codesign_id)
    else:
        print(
            f"Warning: Skipped code signing, env var {CODESIGN_ENV} not set",
            file=sys.stderr,
        )

    dmg_path = build_dmg(
        bundle_source_dir, filename_base, source_dir, dist_dir, product_name
    )

    if notary_user:
        notarize_package(dmg_path, notary_user, notary_password, notary_team_id)
    else:
        print(
            f"Warning: Skipped notarization, env var {NOTARY_USER_ENV} not set",
            file=sys.stderr,
        )


def package_env_vars():
    codesign_id = env.get_env(CODESIGN_ENV, required=False)
    cert_base64 = env.get_env(CERT_P12_ENV, required=False)
    notary_user = env.get_env(NOTARY_USER_ENV, required=False)

    if notary_user:
        notary_password = env.get_env("APPLE_NOTARY_PASSWORD")
        notary_team_id = env.get_env("APPLE_TEAM_ID")
    else:
        notary_password = None
        notary_team_id = None

    if cert_base64:
        cert_password = env.get_env("APPLE_P12_PASSWORD")
    else:
        cert_password = None

    return (
        codesign_id,
        cert_base64,
        cert_password,
        notary_user,
        notary_password,
        notary_team_id,
    )


def build_bundle(bundle_source_dir):

    # it's important to build a new bundle every time, so that we catch bugs with fresh builds.
    if os.path.exists(bundle_source_dir):
        print(f"Bundle already exists, deleting: {bundle_source_dir}")
        shutil.rmtree(bundle_source_dir)

    print("Building bundle...")

    # cmake build install target should run macdeployqt
    cmd_utils.run("cmake --build build --target install", shell=True, print_cmd=True)


def sign_bundle(bundle_source_dir, codesign_id):
    print(f"Signing bundle {bundle_source_dir}...")

    assert_certificate_installed(codesign_id)
    cmd_utils.run(
        [
            CODESIGN_PATH,
            "-f",
            "--options",
            "runtime",
            "--deep",
            "-s",
            codesign_id,
            bundle_source_dir,
        ]
    )


def assert_certificate_installed(codesign_id):
    print(f"Checking certificate: {codesign_id}")

    installed = cmd_utils.run(
        "security find-identity -v -p codesigning",
        get_output=True,
        shell=True,
        print_cmd=True,
    )

    if codesign_id not in installed.stdout:
        raise RuntimeError("Code signing certificate not installed or has expired")


def build_dmg(bundle_source_dir, filename_base, source_dir, dist_dir, product_name):
    settings_path = (
        SETTINGS_FILE if source_dir is None else os.path.join(source_dir, SETTINGS_FILE)
    )
    settings_path_abs = os.path.abspath(settings_path)
    app_path_abs = os.path.abspath(bundle_source_dir)

    # cwd for dmgbuild, since setting the dmg filename to a path (include the dist dir) seems to
    # make the dmg disappear and never writes to the specified path. the dmgbuild module also
    # creates a temporary file in cwd, so it makes sense to change to the dist dir.
    print(f"Changing directory to: {os.path.abspath(dist_dir)}")
    cwd = os.getcwd()
    os.makedirs(dist_dir, exist_ok=True)
    os.chdir(dist_dir)

    try:
        dmg_filename = f"{filename_base}.dmg"
        dmg_path = os.path.join(dist_dir, dmg_filename)
        print(f"Building package {dmg_path}...")
        dmgbuild.build_dmg(
            dmg_filename,
            product_name,
            settings_file=settings_path_abs,
            defines={
                "app": app_path_abs,
            },
        )
    finally:
        print(f"Changing directory back to: {cwd}")
        os.chdir(cwd)

    return dmg_path


def install_certificate(cert_base64, cert_password):
    if not cert_base64:
        raise ValueError("Certificate base 64 not provided")

    if not cert_password:
        raise ValueError("Certificate password not provided")

    with Certificate(cert_base64, "p12") as cert_path:
        print(f"Installing certificate: {cert_path}")

        # WARNING: contains private key password, never print this command
        cmd_utils.run(
            [
                SUDO_PATH,
                SECURITY_PATH,
                "import",
                cert_path,
                "-k",
                KEYCHAIN_PATH,
                "-P",
                cert_password,
                "-T",
                CODESIGN_PATH,
                "-T",
                SECURITY_PATH,
            ],
        )


def notarize_package(dmg_path, user, password, team_id):
    print(f"Notarizing package {dmg_path}...")
    notary_tool = NotaryTool()
    notary_tool.store_credentials(user, password, team_id)
    notary_tool.submit_and_wait(dmg_path)


def get_xcode_path():
    result = cmd_utils.run(
        [XCODE_SELECT_PATH, "-p"], get_output=True, shell=False, print_cmd=True
    )
    return result.stdout.strip()


class NotaryTool:
    """
    Provides a wrapper around the notarytool command line tool.
    """

    def __init__(self):
        self.xcode_path = get_xcode_path()

    def get_path(self):
        return f"{self.xcode_path}{NOTARYTOOL_PATH}"

    def store_credentials(self, user, password, team_id):
        print("Storing credentials for notary tool...")
        notarytool_path = self.get_path()

        # WARNING: contains password, never print this command
        cmd_utils.run(
            [
                notarytool_path,
                "store-credentials",
                "notarytool-password",
                "--team-id",
                team_id,
                "--apple-id",
                user,
                "--password",
                password,
            ]
        )

    def submit_and_wait(self, dmg_filename):
        print("Submitting notarization request...")
        submit_result = self.run_submit_command(dmg_filename)
        request_id = submit_result["id"]

        print(f"Notary submitted, waiting for request: {request_id}")
        start = time.time()
        wait_result = self.run_wait_command(request_id)
        status = wait_result["status"]

        time_taken = time.time() - start
        print(f"Notary complete in {time_taken:.2f}s, status: {status}")
        if status == "Accepted":
            print("Notarization successful.")
        elif status == "Invalid" or status == "Rejected":
            raise ValueError(f"Notarization failed, status: {status}")
        else:
            raise ValueError(f"Unknown status: {status}")

    def run_submit_command(self, dmg_filename):
        if not os.path.exists(dmg_filename):
            raise FileNotFoundError(f"File not found: {dmg_filename}")

        result = cmd_utils.run(
            [
                self.get_path(),
                "submit",
                dmg_filename,
                "--keychain-profile",
                "notarytool-password",
                "--output-format",
                "json",
            ],
            get_output=True,
            shell=False,
            print_cmd=True,
        )

        if result.stderr:
            return json.loads(result.stderr)
        else:
            return json.loads(result.stdout)

    def run_wait_command(self, request_id):
        result = cmd_utils.run(
            [
                self.get_path(),
                "wait",
                request_id,
                "--keychain-profile",
                "notarytool-password",
                "--output-format",
                "json",
            ],
            get_output=True,
            shell=False,
            print_cmd=True,
        )

        if result.stderr:
            return json.loads(result.stderr)
        else:
            return json.loads(result.stdout)



# Deskflow -- mouse and keyboard sharing utility
# Copyright (C) 2024 Symless Ltd.
#
# This package is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# found in the file LICENSE that should have accompanied this file.
#
# This package is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import ctypes, sys, os, shutil, time, subprocess
import xml.etree.ElementTree as ET
import lib.cmd_utils as cmd_utils
import lib.env as env
import psutil  # type: ignore
from lib.certificate import Certificate
import lib.colors as colors
import lib.file_utils as file_utils

LOCK_FILE = "tmp/elevated.lock"
MSBUILD_CMD = "msbuild"
SIGNTOOL_CMD = "signtool"
CERTUTIL_CMD = "certutil"
RUNNER_TEMP_ENV = "RUNNER_TEMP"
SERVICE_NOT_RUNNING_ERROR = 2
ERROR_ACCESS_VIOLATION = 0xC0000005


def run_elevated(script, args=None, use_sys_argv=True, wait_for_exit=False):
    if not args and use_sys_argv:
        args = " ".join(sys.argv[1:])

    if wait_for_exit:
        args += f" --lock-file {LOCK_FILE}"
        env.persist_lock_file(LOCK_FILE)

    command = f"{script} {args} --pause-on-exit"
    print(f"Running script with elevated privileges: {command}")

    WINDOW_HANDLE = None
    OPERATION = "runas"
    DIRECTORY = None
    SHOW_CMD = 1
    instance = ctypes.windll.shell32.ShellExecuteW(
        WINDOW_HANDLE, OPERATION, sys.executable, command, DIRECTORY, SHOW_CMD
    )

    ERROR_ACCESS_DENIED = 5
    if instance == ERROR_ACCESS_DENIED:
        raise RuntimeError(
            f"Failed to run script with elevated privileges, access denied (code {instance})"
        )

    ERROR_MAX = 32
    if instance <= ERROR_MAX:
        raise RuntimeError(
            f"Failed to run script with elevated privileges, error code: {instance}"
        )

    print("Script is running with elevated privileges")

    if wait_for_exit:
        with open(LOCK_FILE, "r") as f:
            pid = f.read()

        print(f"Waiting for elevated process to exit: {pid}")
        while os.path.exists(LOCK_FILE):
            # Intentionally wait forever, since this code should not run where a developer
            # has no control, such as in a CI environment.
            pass


def is_admin():
    """Returns True if the current process has admin privileges."""
    try:
        return ctypes.windll.shell32.IsUserAnAdmin()
    except ctypes.WinError:
        return False


def set_env_var(name, value):
    """
    Sets or updates an environment variable. Appends the value if it doesn't already exist.

    Args:
    name (str): The name of the environment variable.
    value (str): The value of the environment variable.
    """

    current_value = os.getenv(name, "")

    if value not in current_value:
        new_value = f"{current_value}{os.pathsep}{value}" if current_value else value
        os.environ[name] = new_value
        print(f"Setting environment variable: {name}={value}")
        cmd_utils.run(["setx", name, new_value], check=True, shell=True, print_cmd=True)


def package(filename_base, build_dir, dist_dir):
    cert_env_key = "WINDOWS_PFX_CERTIFICATE"
    cert_base64 = env.get_env(cert_env_key, required=False)
    packager = WindowsPackager(filename_base, build_dir, dist_dir)

    if cert_base64:
        cert_password = env.get_env("WINDOWS_PFX_PASSWORD")
        packager.sign_binaries(cert_base64, cert_password)

    packager.build_msi()

    if cert_base64:
        packager.sign_msi(cert_base64, cert_password)
    else:
        print(f"Skipped code signing, env var not set: {cert_env_key}")


def assert_vs_cmd(cmd):
    has_cmd = cmd_utils.has_command(cmd)
    if not has_cmd:
        raise RuntimeError(
            f"The '{cmd}' command was not found, "
            "re-run from 'Developer Command Prompt for VS'"
        )


def run_codesign(path, cert_base64, cert_password):
    time_server = "http://timestamp.digicert.com"
    hashing_algorithm = "SHA256"

    with Certificate(cert_base64, "pfx") as cert_path:
        print("Signing MSI installer...")
        assert_vs_cmd(SIGNTOOL_CMD)

        # WARNING: contains private key password, never print this command
        cmd_utils.run(
            [
                SIGNTOOL_CMD,
                "sign",
                "/f",
                cert_path,
                "/p",
                cert_password,
                "/t",
                time_server,
                "/fd",
                hashing_algorithm,
                path,
            ]
        )


class WindowsPackager:

    def __init__(self, filename_base, build_dir, dist_dir):
        self.filename_base = filename_base
        self.build_dir = build_dir
        self.dist_dir = dist_dir
        self.wix_file = f"{build_dir}/installer/Installer.sln"
        self.msi_file = f"{build_dir}/installer/bin/Release/Installer.msi"

    def build_msi(self):
        print("Building MSI installer...")
        configuration = "Release"
        platform = "x64"

        assert_vs_cmd(MSBUILD_CMD)
        cmd_utils.run(
            [
                MSBUILD_CMD,
                self.wix_file,
                f"/p:Configuration={configuration}",
                f"/p:Platform={platform}",
            ],
            shell=True,
            print_cmd=True,
        )

        path = self.get_package_path()
        print(f"Copying MSI installer to {self.dist_dir}")
        os.makedirs(self.dist_dir, exist_ok=True)
        shutil.copy(self.msi_file, path)

    def get_package_path(self):
        return f"{self.dist_dir}/{self.filename_base}.msi"

    def sign_binaries(self, cert_base64, cert_password):
        exe_pattern = f"{self.build_dir}/bin/*.exe"
        run_codesign(exe_pattern, cert_base64, cert_password)

    def sign_msi(self, cert_base64, cert_password):
        path = self.get_package_path()
        run_codesign(path, cert_base64, cert_password)


class WindowsChoco:
    """Chocolatey for Windows."""

    def ensure_choco_installed(self):
        if cmd_utils.has_command("choco"):
            return

        if not cmd_utils.has_command("winget"):
            print(
                "The winget command was not found, please install Chocolatey manually",
                file=sys.stderr,
            )
            sys.exit(1)

        print("The choco command was not found, installing Chocolatey...")
        cmd_utils.run(
            "winget install chocolatey",
            check=False,
            shell=True,
            print_cmd=True,
        )

        if not cmd_utils.has_command("choco"):
            print(
                "The choco command was still not found, please re-run this script...",
                file=sys.stderr,
            )
            sys.exit(1)


class WindowsService:
    def __init__(self, script, args):
        self.script = script
        self.verbose = args.verbose
        self.bin_name = args.bin_name
        self.source_dir = os.path.abspath(args.source_dir)
        self.target_dir = os.path.abspath(args.target_dir)
        self.service_id = args.service_id
        self.ignore_processes = args.ignore_processes

    def print_verbose(self, message):
        if self.verbose:
            print(message)

    def ensure_admin(self):
        if not is_admin():
            run_elevated(self.script)
            sys.exit()

    def restart(self):
        """Stops the daemon service, copies files, and restarts the daemon service."""

        self.ensure_admin()
        self.stop()
        self.copy_files()
        self.start()

    def reinstall(self):
        """Stops and uninstalls daemon service, copies files, and reinstalls the daemon service."""

        self.ensure_admin()
        self.stop()

        source_bin_path = f"{os.path.join(self.source_dir, self.bin_name)}.exe"

        self.copy_files()

        print("Removing old daemon service")
        try:
            subprocess.run([source_bin_path, "/uninstall"], shell=True, check=True)
        except subprocess.CalledProcessError as e:
            self.check_access_violation(e.returncode, source_bin_path)
            if e.returncode != 0:
                print(
                    f"{colors.WARNING_TEXT} Uninstall failed, return code: {e.returncode}",
                    file=sys.stderr,
                )

        target_bin_path = os.path.join(self.target_dir, self.bin_name + ".exe")

        try:
            print("Installing daemon service")
            subprocess.run([target_bin_path, "/install"], shell=True, check=True)
        except subprocess.CalledProcessError as e:
            self.check_access_violation(e.returncode, target_bin_path)
            if e.returncode != 0:
                print(
                    f"{colors.WARNING_TEXT} Install failed, return code: {e.returncode}"
                )

    def copy_files(self):
        options = file_utils.CopyOptions(ignore_errors=True, verbose=False)
        print(f"Copying files from {self.source_dir} to {self.target_dir}")
        file_utils.copy(f"{self.source_dir}/*", self.target_dir, options)

    def stop(self):
        self.ensure_admin()
        print("Stopping daemon service")
        try:
            subprocess.run(["net", "stop", self.service_id], shell=True, check=True)
        except subprocess.CalledProcessError as e:
            if e.returncode == SERVICE_NOT_RUNNING_ERROR:
                self.print_verbose("Daemon service not running")
            else:
                raise e

        # Wait for Windows to release the file handles after process termination.
        self.wait_for_stop()

    def start(self):
        self.ensure_admin()
        print("Starting daemon service")
        subprocess.run(["net", "start", self.service_id], shell=True, check=True)

    def wait_for_stop(self):
        if self.is_any_process_running(self.target_dir):
            print("Waiting for file handles to release...", end="", flush=True)
            while self.is_any_process_running(self.target_dir):
                if not self.verbose:
                    print(".", end="", flush=True)
                time.sleep(1)
            if not self.verbose:
                print()

    def check_access_violation(self, return_code, bin_path):
        if return_code == ERROR_ACCESS_VIOLATION:
            print(
                f"{colors.WARNING_TEXT} Process crashed with memory access violation: {bin_path}",
                file=sys.stderr,
            )

    def is_ignored_process(self, exe):
        for ignore_process in self.ignore_processes:
            if exe.endswith(ignore_process):
                return True
        return False

    def is_any_process_running(self, dir):
        """Check if there is any running process that contains the given directory."""

        self.print_verbose(f"Checking if any process is running in: {dir}")

        for proc in psutil.process_iter(attrs=["name", "exe"]):
            exe = proc.info["exe"]

            if not exe:
                self.print_verbose(f"Skipping process with no exe: {proc}")
                continue

            if self.is_ignored_process(exe):
                self.print_verbose(f"Ignoring process: {exe}")
                continue

            try:
                if dir.lower() in exe.lower():
                    self.print_verbose(f"Process found: {exe}")
                    return True
            except (psutil.NoSuchProcess, psutil.AccessDenied):
                pass

        return False



# Deskflow -- mouse and keyboard sharing utility
# Copyright (C) 2024 Symless Ltd.
#
# This package is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# found in the file LICENSE that should have accompanied this file.
#
# This package is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import subprocess
import sys
import lib.env as env

try:
    import colorama  # type: ignore
    from colorama import Fore  # type: ignore

    colorama.init()
except ImportError:

    class Fore:
        RESET = ""
        YELLOW = ""


def has_command(command):
    platform = sys.platform
    if platform == "win32":
        cmd = f"where {command}"
    else:
        cmd = f"which {command}"
    try:
        subprocess.check_output(cmd, shell=True)
        return True
    except subprocess.CalledProcessError:
        return False


def strip_continuation_sequences(command, strip_newlines=True):
    """
    Remove the continuation sequences (\\) from a command.

    To spread strings over multiple lines in YAML files, like in bash, a backslash is used at
    the end of each line as continuation character.
    """

    if isinstance(command, list):
        raise ValueError("List commands are not supported")

    cmd_continuation = " \\"
    command = command.replace(cmd_continuation, "")

    # Some versions of pyyaml will remove the newlines already, so always stripping
    # makes the output more consistent.
    if strip_newlines:
        command = command.replace("\n", " ")

    return command


def run(
    command,
    check=True,  # true by default to fail fast
    shell=False,  # false by default for security
    get_output=False,
    print_cmd=False,  # false by default for security
):
    """
    Convenience wrapper around `subprocess.run` to:
    - print the command before running it (if `print_cmd` is True)

    This differs to `subprocess.run` in that by default it:
    - checks the return code by default
    - prints list commands as a readable string on failure

    This is the same as `subprocess.run` in that it:
    - does not use shell by default for security (shell is less secure)

    Args:
        command (str or list): The command to run.
        check (bool): Raise an exception if the command fails.
        shell (bool): Run the command in a shell (false by default for security)
        get_output (bool): Return the output of the command.
        print_cmd (bool): Print the command before running it (false by default for security)
    """

    is_list_cmd = isinstance(command, list)

    # create string version of list command, only for debugging purposes
    command_str = command
    if is_list_cmd:
        command_str = " ".join(command)

    if print_cmd:
        print(f"Running: {command_str}")
    else:
        print("Running command...")
        command_str = "***"

    # TODO: You can definitely use a list command with shell=True on Windows,
    # but can you use a string command with shell=False on Windows?
    #
    # The `subprocess.run` function has a little gotcha:
    # - a string command must be used when `shell=True`
    # - a list command must be used when shell isn't or `shell=False`
    # however, it allows you to pass a string command when shell isn't used or `shell=False`
    # then fails with a vague error message. same problem with list commands and `shell=True`
    if not env.is_windows() and is_list_cmd and shell:
        raise ValueError("List commands cannot be used when shell=True on Unix systems")
    elif not is_list_cmd and not shell:
        raise ValueError("String commands cannot be used when shell=False or not set")

    # Flush the output to ensure the command is printed before the output of the command,
    # which seems to happen in the GitHub runner logs.
    sys.stdout.flush()
    sys.stderr.flush()

    try:
        if get_output:
            result = subprocess.run(
                command,
                shell=shell,
                check=check,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True,
            )
        else:
            result = subprocess.run(command, check=check, shell=shell)

    except subprocess.CalledProcessError as e:
        # Take control of how failed commands are printed:
        # - if `print_cmd` is false, it will print `***` instead of the command
        # - if the command was a list, the command is printed as a readable string
        raise RuntimeError(
            f"Command exited with code {e.returncode}: {command_str}"
        ) from None
    except Exception:
        # Take control of how failed commands are printed:
        # - if `print_cmd` is false, it will print `***` instead of the command
        # - if the command was a list, the command is printed as a readable string
        raise RuntimeError(f"Command failed: {command_str}")

    if result.returncode != 0:
        print(
            f"{Fore.YELLOW}Command exited with code {result.returncode}:{Fore.RESET} {command_str}",
            file=sys.stderr,
        )

    return result
